@@ -23,12 +23,16 @@
  * Portions adapted from gnome-shell/src/shell-global.c
  */
 
+/**
+ * SECTION:meta-background-actor
+ * @title: MetaBackgroundActor
+ * @short_description: Actor for painting the root window background
+ */
+
 #include <config.h>
 
-#define COGL_ENABLE_EXPERIMENTAL_API
 #include <cogl/cogl-texture-pixmap-x11.h>
 
-#define CLUTTER_ENABLE_EXPERIMENTAL_API
 #include <clutter/clutter.h>
 
 #include <X11/Xatom.h>
